# Copyright 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import print_function

import greenlet
import eventlet
import eventlet.pools
import socket
import json
from eventlet import greenthread
from collections import namedtuple
from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import (
    DockerContainer,
    DockerPortMap,
    DockerSwarmCluster,
    DockerSwarmNode,
    DockerNetwork,
    DockerService,
    DockerHost
)

#
# Import docker python library 
try:
    import docker
    DOCKER_INSTALLED = True
except ImportError as e:
    DOCKER_INSTALLED = False

IpMacTuple = namedtuple('IpMacTuple', ['network_name', 'network_id', 'ip', 'mac', 'gw', 'prefix_len'])
PortMapTuple = namedtuple('PortMapTuple', ['proto', 'port', 'host_ip', 'host_port'])

class DockerMonitor(object):
    """Class VmwareHypervisor
    Interfaces with VCenter. It provides method to launch a thread per vcenter
    and update database with VMware VM/ESX information

    In future this class can be extended to interface with vRealize or vCloudSuite etc
    """
    #Default Docker Polling Period
    _DOCKER_DEFAULT_POLL_PERIOD = 30
    _DOCKER_MIN_POLL_PERIOD = 10

    def parse_config(self, config):
        '''
        Validate user hasn't specified any parameter that
        netq agent does not understand
        '''
        valid_keys = set(['poll_period', 'enable'])
        self.__enabled = False

        for key in config.keys():
            if key not in valid_keys:
                self.__logger.error("Invalid configuration: %r" % key)
                return False

        self._poll_period = config.get('poll_period',
                                       DockerMonitor._DOCKER_DEFAULT_POLL_PERIOD)

        if self._poll_period < DockerMonitor._DOCKER_MIN_POLL_PERIOD:
            self.__logger.error("Configured poll_period:%r Support min poll_period:%r. Applying min poll" %
                                (self._poll_period, DockerMonitor._DOCKER_MIN_POLL_PERIOD))
            self._poll_period = DockerMonitor._DOCKER_MIN_POLL_PERIOD

        self.__enabled = config.get('enable', False)
        return

    def __init__(self, yaml_config, logger, platform=None, pool=None):
        self.__logger = logger
        self.__running = False
        self.__enabled = False
        self.__monitoring_events = False
        self.__client = None
        self._host = socket.gethostname()
        self._gt = None
        self._event_handler_gt = None
        self.__pool = None

        # 1. Check if docker monitoring is supported on the platform
        if not platform:
            self.__logger.info("Platform is not defined. Skip Docker Monitoring...")
            return

        if 'ubuntu' not in platform.lower():
            self.__logger.info("Docker Monitoring not Supported on %r..." % platform)
            return

        if not DOCKER_INSTALLED:
            self.__logger.info("Docker is not installed on %r..." % platform)
            return

        # 2. Check if docker monitoring is configured and enabled
        if yaml_config and 'docker' in yaml_config:
            config = yaml_config['docker']
            self.parse_config(config)

        if not self.__enabled:
            self.__logger.info("Docker Monitoring not enabled...")
            return

        self.__pool = pool or eventlet.GreenPool()

        try:
            self.__client = docker.DockerClient(base_url='unix://var/run/docker.sock')
        except Exception as e:
            self.__logger.error("Failed to initialize DockerClient: %r" % e)
        return

    @property
    def poll_period(self):
        return self._poll_period

    @property
    def enabled(self):
        return self.__enabled

    @property
    def running(self):
        return self.__running

    @staticmethod
    def parse_port_mapping(container):
        '''
        Description:
        This function prases container config and returns container
        ports to host ports mapping list

        Input:
        @container: container config dictionary returned by docker.sock

        Return:
        List of (port, host_ip, host_port) namedtuple
        '''
        net_settings = container.attrs.get('NetworkSettings', {})
        ports = net_settings.get('Ports', {})
        port_map_list = []
        if ports:
            for port, host_ports in ports.iteritems():
                (port, proto) = str(port).split('/')

                if not host_ports:
                    portmap = PortMapTuple(proto, port, '*', '')
                    port_map_list.append(portmap)
                    continue

                for host_map in host_ports:
                    host_ip = str(host_map.get('HostIp', '*'))
                    host_port = str(host_map.get('HostPort', ''))
                    portmap = PortMapTuple(proto, port, host_ip, host_port)
                    port_map_list.append(portmap)
        return port_map_list

    @staticmethod
    def parse_network_config(container):
        '''
        Description:
        This function prases NetworkSettings/Networks config and returns
        (ip, mac) addresses assigned to the container

        Input:
        @container: container config dictionary returned by docker.sock

        Return:
        List of (ip, mac) namedtuple
        '''
        ip_mac_list = []
        network_settings = container.attrs.get('NetworkSettings', {})
        networks = network_settings.get('Networks', {})
        for network_name, network in networks.iteritems():
            ip = str(network.get('IPAddress', ''))
            mac = str(network.get('MacAddress', ''))
            net_id = str(network.get('NetworkID', ''))
            gw = str(network.get('Gateway', ''))
            prefix_len = str(network.get('IPPrefixLen', ''))
            ipmac = IpMacTuple(network_name.encode('ascii','ignore'), net_id, ip, mac, gw, prefix_len)
            ip_mac_list.append(ipmac)
        return ip_mac_list

    @staticmethod
    def parse_container_labels(labels, logger=None):
        '''
        This function parses container labels dictionary and returns
        (service_id,service_name)
        '''
        service_id = ''
        service_name = ''

        try:
            for label, value in labels.iteritems():
                if str(label) == 'com.docker.swarm.service.id':
                    service_id = str(value)

                if str(label) == 'com.docker.swarm.service.name':
                    service_name = str(value)

        except Exception as e:
            if logger:
                logger.error("Failed to parse Container Labels: %r" % e)
            pass

        return (service_id, service_name)

    def delete_container_info(self, container_id):
        """
        Delete a container from database
        """
        m_DockerContainer = DockerContainer.memmodel()

        self.__logger.info('Function:delete_container_info id:%r... ' % (container_id))
        for entry in m_DockerContainer.query.filter(hostname=HOSTNAME, container_id=container_id):
            old_entry = DockerContainer()
            old_entry.copy(entry)
            self.__logger.info('Deleting id:%r from database ' % (container_id))
            old_entry.delete()
        return

    def delete_container_portmap(self, container_name):
        """
        Delete a container portmap
        """
        m_DockerPortMap = DockerPortMap.memmodel()

        self.__logger.info('Function:delete_container_portmap id:%r... ' % (container_name))
        for entry in m_DockerPortMap.query.filter(hostname=HOSTNAME, name=container_name):
            old_entry = DockerPortMap()
            old_entry.copy(entry)
            self.__logger.info('Deleting id:%r from container_portmap database ' % (container_name))
            old_entry.delete()
        return


    def delete_docker_network(self, network_id):
        m_DockerNetwork = DockerNetwork.memmodel()
        self.__logger.info('Function:delete_docker_network id:%r... ' % (network_id))

        for entry in m_DockerNetwork.query.filter(hostname=HOSTNAME, network_id=network_id):
            old_entry = DockerNetwork()
            old_entry.copy(entry)
            self.__logger.info('Deleting network:%r from database ' % (network_id))
            old_entry.delete()
        return


    def flush_container_info(self, container_list):
        m_DockerContainer = DockerContainer.memmodel()
        self.__logger.info('Flush stale container info from db... ')

        for entry in m_DockerContainer.query.filter(hostname=HOSTNAME):
            if (entry.container_id) in container_list:
                continue
            old_entry = DockerContainer()
            old_entry.copy(entry)
            self.__logger.info('Deleting Container id:%r from database ' % (entry.container_id))
            old_entry.delete()
        return


    def flush_container_portmap(self, container_list):
        m_DockerPortMap = DockerPortMap.memmodel()
        self.__logger.info('Flush stale container portmap from db... ')

        for entry in m_DockerPortMap.query.filter(hostname=HOSTNAME):
            key = (entry.container_ip, entry.proto, entry.container_port,
                   entry.host_ip, entry.host_port)

            if key in container_list:
                continue

            old_entry = DockerPortMap()
            old_entry.copy(entry)
            self.__logger.info('Deleting Portmap id:(%r,%r,%r) from database ' %
                               (entry.container_ip, entry.proto, entry.container_port))
            old_entry.delete()
        return


    def flush_docker_network(self, network_list):
        m_DockerNetwork = DockerNetwork.memmodel()
        self.__logger.info('Flush stale docker network from db... ')

        for entry in m_DockerNetwork.query.filter(hostname=HOSTNAME):
            key = (entry.network_name, entry.subnet)

            if key in network_list:
                continue

            old_entry = DockerPortMap()
            old_entry.copy(entry)
            self.__logger.info('Deleting Network: %r subnet: %r from database ' %
                               (entry.network_name, entry.subnet))
            old_entry.delete()
        return


    def flush_docker_swarm_cluster(self, cluster_list):
        m_DockerSwarmCluster = DockerSwarmCluster.memmodel()
        self.__logger.info('Flush stale docker swarm cluster from db... ')

        for entry in m_DockerSwarmCluster.query.filter(hostname=HOSTNAME):
            key = (entry.cluster_name)

            if key in cluster_list:
                continue

            old_entry = DockerPortMap()
            old_entry.copy(entry)
            self.__logger.info('Deleting Swarm Cluster: %r from database ' %
                               entry.cluster_name)
            old_entry.delete()
        return


    def flush_docker_swarm_node(self, node_list):
        m_DockerSwarmNode = DockerSwarmNode.memmodel()
        self.__logger.info('Flush stale docker swarm node from db... ')

        for entry in m_DockerSwarmNode.query.filter(hostname=HOSTNAME):
            key = (entry.cluster_id, entry.node_id)

            if key in node_list:
                continue

            old_entry = DockerSwarmNode()
            old_entry.copy(entry)
            self.__logger.info('Deleting Swarm Node: %r node: %r from database' %
                               (entry.cluster_id, entry.node_id))
            old_entry.delete()
        return


    def flush_docker_swarm_service(self, service_list):
        m_DockerService = DockerService.memmodel()
        self.__logger.info('Flush stale docker swarm service from db... ')

        for entry in m_DockerService.query.filter(hostname=HOSTNAME):
            if (entry.service_name) in service_list:
                continue

            old_entry = DockerService()
            old_entry.copy(entry)
            self.__logger.info('Deleting Swarm swarm: %r' % (entry.service_name))
            old_entry.delete()
        return

    def update_container_info(self, container):
        update_list = []

        config = container.attrs.get('Config', {})
        docker_container = DockerContainer()
        container_name = str(container.attrs.get('Name', ''))
        """
        Hack: Docker API returns name with leading '/'
        """
        if container_name[0] == '/':
            container_name = container_name[1:]

        self.__logger.info('Updating container:%r info' %
                               (docker_container.name))
        docker_container.name = container_name
        docker_container.container_id = str(container.attrs.get('Id', ''))
        docker_container.image = str(config.get('Image',''))
        state = container.attrs.get('State', {})
        labels = config.get('Labels', {})
        (service_id, service_name) = DockerMonitor.parse_container_labels(labels, self.__logger)
        docker_container.service_id = service_id
        docker_container.service_name = service_name
        docker_container.status = str(state.get('Status', ''))
        docker_container.start_time = state.get('StartedAt','').encode('ascii', 'ignore')

        ip_mac_list = DockerMonitor.parse_network_config(container)
        port_map_list = DockerMonitor.parse_port_mapping(container)
        port_list = []
        for portmap in port_map_list:
            port_list.append((portmap.proto, portmap.port))

        docker_container.port_list = str(port_list)

        for ipmac in ip_mac_list:
            docker_container.ip =  ipmac.ip
            docker_container.mac = ipmac.mac
            docker_container.gw = ipmac.gw
            docker_container.prefix_len = ipmac.prefix_len
            docker_container.network_name = ipmac.network_name
            docker_container.network_id = ipmac.network_id

            docker_container.save()
            update_list.append(docker_container.container_id)
            self.__logger.info('Docker container:%r (%r)info pushed to db' %
                               (docker_container.name, docker_container.ip))
        if not ip_mac_list:
            docker_container.save()
            update_list.append(docker_container.container_id)
            self.__logger.info('Docker container:%r (%r)info without IP/MAC pushed to db' %
                               (docker_container.name, docker_container.ip))

        return update_list


    def update_container_port_mapping(self, container):
        update_list = []

        try:
            info = self.__client.info()
        except Exception as e:
            self.__logger.error("Failed to get Docker Client information: %r" % e)
            return update_list

        docker_port = DockerPortMap()
        swarm_info = info.get('Swarm', {})
        swarm_node_id = str(swarm_info.get('NodeID', ''))
        docker_port.node_id = swarm_node_id
        config = container.attrs.get('Config', {})
        ip_mac_list = DockerMonitor.parse_network_config(container)
        port_map_list = DockerMonitor.parse_port_mapping(container)

        docker_port.container_id = str(container.attrs.get('Id', ''))
        container_name = str(container.attrs.get('Name', ''))
        """
        Hack: Docker API returns name with leading '/'
        """
        if container_name[0] == '/':
            container_name = container_name[1:]
        docker_port.name = container_name
        docker_port.image = str(config.get('Image',''))

        # Create Container port to host mapping.  
        # If container or host has multiple IPs 
        # create a mapping for each IP address 
        for ipmac in ip_mac_list:
            docker_port.network_name = ipmac.network_name
            docker_port.network_id = ipmac.network_id
            docker_port.container_ip = ipmac.ip
            docker_port.mac = ipmac.mac

            if ipmac.network_name == 'host':
                """ If container is attached to host network
                    Check for exposed ports
                """
                exposed_port_list = config.get('ExposedPorts', {})
                for exposed_port in exposed_port_list.keys():
                    if '/' in exposed_port:
                        (port, proto) = str(exposed_port).split('/')
                        docker_port.proto = proto
                        docker_port.container_port = port
                        docker_port.host_port = port
                        docker_port.host_ip = '*'
                        self.__logger.info('Docker container:%r port:(%r/%r) pushed to db' %
                                            (docker_port.name, docker_port.proto,
                                            docker_port.container_port))
                        docker_port.save()
                        update_list.append((docker_port.container_ip,
                                            docker_port.proto, docker_port.container_port,
                                            docker_port.host_ip, docker_port.host_port))
            else:
                for portmap in port_map_list:
                    docker_port.proto = portmap.proto
                    docker_port.container_port = portmap.port
                    docker_port.host_port = portmap.host_port
                    docker_port.host_ip = portmap.host_ip
                    self.__logger.info('Docker container:%r port:(%r/%r) host(%r:%r)pushed to db' %
                                       (docker_port.name, docker_port.proto,
                                       docker_port.container_port, docker_port.host_ip,
                                       docker_port.host_port))
                    docker_port.save()
                    update_list.append((docker_port.container_ip,
                                        docker_port.proto, docker_port.container_port,
                                        docker_port.host_ip, docker_port.host_port))

        return update_list


    def update_docker_network(self, network):
        update_list = []

        docker_network = DockerNetwork()

        network_attrs = network.attrs
        docker_network.driver = str(network_attrs.get('Driver', ''))
        docker_network.ipv6_enabled = network_attrs.get('EnableIPv6', False)
        docker_network.network_name = str(network_attrs.get('Name', ''))
        docker_network.network_id = str(network_attrs.get('Id', ''))
        options = network_attrs.get('Options', {})

        for option_type, option_value in options.iteritems():
            option_type = str(option_type)

            if option_type == 'com.docker.network.bridge.default_bridge':
                docker_network.default_bridge = bool(option_value)

            if option_type == 'com.docker.network.bridge.host_binding_ipv4':
                docker_network.host_binding = str(option_value)

            if option_type == 'com.docker.network.driver.overlay.vxlanid_list':
                docker_network.vxlan_id = int(option_value)

            if option_type == 'encrypted':
                docker_network.encrypted = True

            if option_type == 'com.docker.network.bridge.enable_ip_masquerade':
                docker_network.ip_masquerade = bool(option_value)

            if option_type == 'com.docker.network.driver.mtu':
                docker_network.mtu = int(option_value)

        containers = network.attrs.get('Containers', {})
        container_list = []
        container_string = None
        if containers:
            for container_id, container_info in containers.iteritems():
                ipv4_addr = str(container_info.get('IPv4Address', ''))
                ipv6_addr = str(container_info.get('IPv6Address', ''))
                name = str(container_info.get('Name',''))
                if container_string:
                    container_string = container_string + ', '

                if ipv6_addr == '' and ipv4_addr != '' :
                    container_string = 'Name:%s IPv4:%s' % (name, ipv4_addr)
                elif ipv6_addr != '' and ipv4_addr != '':
                    container_string = 'Name:%s IPv4:%s IPv6:%s' % (name, ipv4_addr, ipv6_addr)
                elif ipv6_addr != '':
                    container_string = 'Name:%s IPv4:%s IPv6:%s' % (name, ipv4_addr, ipv6_addr)
                else:
                    container_string = 'Name:%s' % (name)


                container_list.append(container_string)

        docker_network.container_list = ', '.join(container_list)
        ipam = network_attrs.get('IPAM', {})
        ipam_config_list = ipam.get('Config', [])

        if not ipam_config_list:
            docker_network.save()
            update_list.append((docker_network.network_name, docker_network.subnet))
        else:
            for ipam_config in ipam_config_list:
                docker_network.subnet = str(ipam_config.get('Subnet',''))
                docker_network.gateway = str(ipam_config.get('Gateway',''))
                docker_network.ipam_driver = str(ipam_config.get('Driver',''))
                self.__logger.info('Docker Network:%r (%r)info pushed to db' %
                                   (docker_network.network_name, docker_network.subnet))
                docker_network.save()
                update_list.append((docker_network.network_name, docker_network.subnet))

        return update_list


    def update_docker_swarm_service(self, service):
        update_list = []

        docker_service = DockerService()
        spec = service.attrs.get('Spec', {})
        docker_service.service_name = str(spec.get('Name', ''))
        update_config = spec.get('UpdateConfig', {})
        docker_service.parallelism = int(update_config.get('Parallelism', 0))
        docker_service.service_id = str(service.attrs.get('ID', ''))
        mode = spec.get('Mode', {})
        task_template = spec.get('TaskTemplate', {})
        container_spec = task_template.get('ContainerSpec', {})
        image = container_spec.get('Image', '')
        docker_service.image = str(image)
        for mode_type, mode_value in mode.iteritems():
            mode_type = str(mode_type)
            docker_service.mode = mode_type
            if mode_type == 'Replicated':
                docker_service.replicas = mode_value.get('Replicas', 0)

        network_list = []
        vip_list = []
        endpoint = service.attrs.get('Endpoint', {})
        vips = endpoint.get('VirtualIPs', [])
        for vip in vips:
            network_id = str(vip.get('NetworkID', ''))
            networks =  self.__client.networks.list(ids=[network_id])
            for network in networks: 
                network_list.append(str(network.attrs.get('Name','')))
            vip = str(vip.get('Addr' ''))
            vip_list.append(vip)
        docker_service.network_list = ', '.join(network_list)
        docker_service.vip = ', '.join(vip_list)
        # Port 
        endpoint_spec = endpoint.get('Spec', {})
        ports = endpoint_spec.get('Ports', {})
        port_list = []
        for port in ports:
            proto = str(port.get('Protocol',''))
            published_port = port.get('PublishedPort', 0)
            target_port = port.get('TargetPort', 0)
            port_str = '%s:%s:%s' % (proto, published_port, target_port)
            port_list.append(port_str)
        docker_service.port_list = ', '.join(port_list)
        version = service.attrs.get('Version',{})
        docker_service.version = int(version.get('Index', 0))
        self.__logger.info('Docker Swarm Service :%r info pushed to db' %
                               (docker_service.service_name))
        docker_service.save()
        update_list.append(docker_service.service_name)

        return update_list


    def poll_docker_host(self):
        try:
           info = self.__client.info()
        except Exception as e:
            self.__logger.error("Failed to get Docker Client information: %r" % e)
            return

        docker_host = DockerHost()
        docker_host.docker_version = str(info.get('ServerVersion', ''))
        docker_host.containers = info.get('Containers', 0)
        docker_host.containers_running = info.get('ContainersRunning', 0)
        docker_host.images = info.get('Images',0)
        docker_host.ip_forwarding = info.get('IPv4Forwarding',False)

        self.__logger.info('Updating Docker Host info: Num containers:%r, images:%r' %
                          (docker_host.containers, docker_host.images))
        docker_host.save()
        return


    def poll_docker_network(self):
        update_list = []
        try:
            network_list = self.__client.networks.list()
        except Exception as e:
            self.__logger.error("Failed to get Docker Network list: %r" % e)
            return

        self.__logger.info('Updating Docker Network info')
        for network in network_list:
            update_list = update_list + self.update_docker_network(network)

        # remove any stale entries from database 
        self.flush_docker_network(update_list)
        return


    def poll_docker_info(self):
        portmap_update_list = []
        container_update_list = []
        try:
            container_list = self.__client.containers.list()
        except Exception as e:
            self.__logger.error("Failed to get Docker container list: %r" % e)
            return


        self.__logger.info('Updating Docker Container and portmap info')


        for container in container_list:
            container_update_list = container_update_list + self.update_container_info(container)
            portmap_update_list = portmap_update_list + self.update_container_port_mapping(container)

        # Remove stale entries from database
        self.flush_container_info(container_update_list)
        self.flush_container_portmap(portmap_update_list)
        return


    def poll_docker_swarm_cluster(self):
        update_list = []
        try:
            info = self.__client.info()
        except Exception as e:
            self.__logger.error("Failed to get Docker Client information: %r" % e)
            return

        try:
            nodes = self.__client.nodes.list()
        except Exception as e:
            self.__logger.info("Failed to get Docker Swarm Cluster nodes: %r" % e)
            return

        node_dict = {}
        node_list = []
        for node in nodes:
            node_id = node.attrs.get('ID', None)
            if not node_id:
                continue
            node_description = node.attrs.get('Description', {})
            node_name = str(node_description.get('Hostname', ''))
            node_dict[str(node_id)] = node_name
            node_list.append(node_name)

        swarm_info = info.get('Swarm', {})
        swarm_enabled = swarm_info.get('ControlAvailable', False)
        self.__logger.info("Docker Swarm State: %r" % swarm_enabled)

        if swarm_enabled:
            swarm_cluster_info = swarm_info.get('Cluster', {})
            cluster_spec = swarm_cluster_info.get('Spec', {})
            docker_swarm = DockerSwarmCluster()
            docker_swarm.docker_version = str(info.get('ServerVersion', ''))
            docker_swarm.cluster_id = str(swarm_cluster_info.get('ID', ''))
            docker_swarm.cluster_name = str(cluster_spec.get('Name', ''))
            docker_swarm.cluster_version = str(cluster_spec.get('Version', ''))
            docker_swarm.num_nodes = swarm_info.get('Nodes', 0)
            docker_swarm.num_managers = swarm_info.get('Managers', 0)
            remote_managers =  swarm_info.get('RemoteManagers', {})
            manager_list = []
            for remote_manager in remote_managers:
                manager_addr = str(remote_manager.get('Addr', ''))
                node_id = str(remote_manager.get('NodeID', ''))
                node_name = node_dict.get(node_id, '')
                remote_manager_str = "%s:%s" % (manager_addr, node_name)
                manager_list.append(remote_manager_str)
            docker_swarm.managers = ', '.join(manager_list)
            docker_swarm.nodes = ', '.join(node_list)
            self.__logger.info('Docker swarm:%r info pushed to db' %
                                   docker_swarm.cluster_name)
            docker_swarm.save()
            update_list.append(docker_swarm.cluster_name)

        # Remove any stale entries from db
        self.flush_docker_swarm_cluster(update_list)

        return


    def poll_docker_swarm_node(self):
        update_list = []
        try:
            nodes = self.__client.nodes.list()
        except Exception as e:
            self.__logger.info("Failed to get Docker Swarm node information: %r" % e)
            return

        try:
            info = self.__client.info()
        except Exception as e:
            self.__logger.error("Failed to get Docker Client information: %r" % e)
            return

        swarm_info = info.get('Swarm', {})
        swarm_cluster_info = swarm_info.get('Cluster', {})
        cluster_spec = swarm_cluster_info.get('Spec', {})
        cluster_id = str(swarm_cluster_info.get('ID', ''))
        cluster_name = str(cluster_spec.get('Name',''))

        docker_swarm_node = DockerSwarmNode()
        docker_swarm_node.cluster_id = cluster_id
        docker_swarm_node.cluster_name = cluster_name

        self.__logger.info("Updating Docker Swarm %r Nodes..." % docker_swarm_node.cluster_name)
        for node in nodes:
            description = node.attrs.get('Description', {})
            spec = node.attrs.get('Spec', {})
            engine = description.get('Engine', {})
            status = node.attrs.get('Status', {})
            state = str(status.get('State', {}))
            docker_swarm_node.docker_version = str(engine.get('EngineVersion', ''))
            docker_swarm_node.role = str(spec.get('Role', ''))
            docker_swarm_node.node_id = str(node.attrs.get('ID', ''))
            docker_swarm_node.node_name = str(description.get('Hostname',''))
            docker_swarm_node.node_state = state
            plugin_list = engine.get('Plugins', [])
            plugin_dict = {}
            for plugin in plugin_list:
                plugin_type = str(plugin.get('Type', ''))
                plugin_name = str(plugin.get('Name', ''))
                if plugin_type in plugin_dict:
                    plugin_dict[plugin_type].append(plugin_name)
                else:
                    plugin_dict[plugin_type] = [plugin_name]
            for key, value in plugin_dict.items():
                plugin_dict[key] = ', '.join(value)
            plugin_string = ', '.join(['%s:(%s)' % (k, v) for (k, v) in plugin_dict.items()])
            docker_swarm_node.plugins = plugin_string
            self.__logger.info('Updating Docker swarm node %r info' % docker_swarm_node.node_id)
            docker_swarm_node.save()
            update_list.append((docker_swarm_node.cluster_id, docker_swarm_node.node_id))

        self.flush_docker_swarm_node(update_list)
        return


    def poll_docker_swarm_service(self):
        update_list = []
        try:
            service_list = self.__client.services.list()
        except Exception as e:
            self.__logger.info("Failed to get Docker Services list: %r" % e)
            return

        self.__logger.info('Updating Docker swarm Services info')
        for service in service_list:
            update_list = update_list + self.update_docker_swarm_service(service)

        self.flush_docker_swarm_service(update_list)
        return


    def _run(self):
        """
        Periodically monitor docker engine
        """
        self.__running = True

        while True:
            self.poll_docker_host()
            self.poll_docker_info()
            self.poll_docker_network()
            self.poll_docker_swarm_cluster()
            self.poll_docker_swarm_node()
            self.poll_docker_swarm_service()
            greenthread.sleep(self.poll_period)

        self.__logger.info('Poll loop terminated!!')
        return


    def handle_container_event(self, action, actor):
        container_id = actor.get('ID').encode('ascii', 'ignore')
        attributes = actor.get('Attributes', {})
        container_name = attributes.get('name').encode('ascii', 'ignore')
        self.__logger.info('handle_container_event %r(%r) action: %r ' % (container_id, container_name, action))

        if (action == 'kill') or (action == 'die'):
            self.__logger.info("Deleting container: %r(%r) from db" % (container_id, container_name))
            self.delete_container_info(container_id)
            self.delete_container_portmap(container_name)

        elif (action == 'create') or (action == 'start'):
            container_list = self.__client.containers.list(filters={'id':container_id})
            if not container_list:
                self.__logger.info("handle_container_event couldn't find container: %r" % (container_id))

            for container in container_list:
                self.update_container_info(container)
                self.update_container_port_mapping(container)
        else:
           self.__logger.info('Skipping handle_container_event action: %r' % (action))

        return


    def handle_network_event(self, action, actor):
        network_id = actor.get('ID').encode('ascii', 'ignore')
        self.__logger.info('handle_network_event %r action: %r' % (network_id, action))

        if ((action == 'create') or
            (action == 'connect') or
            (action == 'disconnect')):

            network_list = self.__client.networks.list(filters={'id':network_id})
            if not network_list:
               self.__logger.info('Failed to find %r action: %r' % (network_id, action))

            for network in network_list:
                self.update_docker_network(network)

        elif (action == 'destroy') :
            self.delete_docker_network(network_id)
        else:
           self.__logger.info('Skipping handle_network_event action: %r' % (action))
        return


    def _event_handler(self):
        """
        Handler docker engine events
        """
        self.__monitoring_events = True

        self.__logger.info('Docker Event Monitoring started...')
        for event in self.__client.events():
            self.__logger.info('Docker event Event received...')
            event_dict = json.loads(event)
            event_type = event_dict.get('Type').encode('ascii', 'ignore')
            action = event_dict.get('Action').encode('ascii', 'ignore')
            actor = event_dict.get('Actor', {})
            self.__logger.info('Docker %r %r event received' % (event_type, action))

            if event_type == 'container':
                self.handle_container_event(action, actor)
                self.poll_docker_host()
            elif event_type == 'network':
                self.handle_network_event(action, actor)
            elif event_type == 'image':
                self.poll_docker_host()
            else:
                self.__logger.error('Skipping %r event' % (event_type))
        return


    def start(self):
        """Start Monitor docker thread
        """
        # If docker monitoring isn't enabled skip
        if not self.__enabled:
            self.__logger.info('Docker monitor is not enabled ')
            return

        # Check docker client connection is initialized 
        if not self.__client:
            self.__logger.error('Docker Monitor not initialized...')
            return

        if not self.__monitoring_events:
            self.__logger.info('Starting Docker Event Monitor thread...')
            self._event_handler_gt = self.__pool.spawn(self._event_handler)
            self._event_handler_gt.link(self.stop_monitor_checker)
        else:
            self.__logger.info('Docker Event Monitor is already running...')

        if not self.__running:
            self.__logger.info('Starting Docker Poll thread...')
            self._gt = self.__pool.spawn(self._run)
            self._gt.link(self.stop_poll_checker)
        else:
            self.__logger.info('Docker Poll thread is already running...')
        return


    def stop_poll_checker(self, *args, **kwargs):
        """
        stop_poll_checker()
        should be called when docker poll thread is stopped/killed
        """
        if self.__running:
            self.__logger.info('Docker Poll thread stopped')
            self.__running = False
        return


    def stop_monitor_checker(self, *args, **kwargs):
        """
        stop_monitor_checker()
        should be called when docker monitor event thread is stopped/killed
        """
        if self.__monitoring_events:
            self.__logger.info('Docker Monitor event thread stopped')
            self.__monitoring_events = False
        return


    def halt(self):
        if self._gt:
            self.__logger.info('Terminating docker monitor: %r' % self._gt)
            eventlet.kill(self._gt, eventlet.StopServe)
        self._gt = None
        return
