# Copyright 2017 Cumulus Networks, Inc.  All Rights reserved.
#
# This software is subject to the Cumulus Networks End User License Agreement
# available at the following locations:
#
# Internet: https://cumulusnetworks.com/downloads/eula/latest/view/
# Cumulus Linux systems: /usr/share/cumulus/EULA.txt

from __future__ import absolute_import

import json

from netq_lib.orm import HOSTNAME
from netq_lib.orm.redisdb.models import BgpSession
from netq_lib.common.enums import ServiceStatus


def process_bgp_neighbors(output, svc_failed=False, **kwargs):
    """Build summary of BGP node info, such as ASN & RID
    """
    retcode = ServiceStatus.OK

    try:
        json_out = json.loads(output)
    except ValueError:
        # We treat invalid JSON as the infomation terminating
        json_out = []
        retcode = ServiceStatus.FAIL

    # first process the entries already in the DB
    m_bgp_session = BgpSession.memmodel()
    for entry in list(m_bgp_session.query.filter(hostname=HOSTNAME)):
        if entry.state == 'Fail':
            continue
        if entry.vrf == 'default':
            vrf = 'Default'
        else:
            vrf = entry.vrf
        old_bgp_entry = BgpSession()
        if (vrf not in json_out or
                entry.peer_name not in json_out[vrf]):
            if not svc_failed:
                old_bgp_entry.copy(entry)
                old_bgp_entry.delete()
            else:
                old_bgp_entry.state = 'Fail'
                old_bgp_entry.vrf = entry.vrf
                old_bgp_entry.asn = entry.asn
                old_bgp_entry.peer_name = entry.peer_name
                old_bgp_entry.save()
        else:
            # Check that ASN hasn't changed
            if json_out[vrf][entry.peer_name]['localAs'] != entry.asn:
                old_bgp_entry.copy(entry)
                old_bgp_entry.delete()

    for vrf in json_out:
        vrf_id = json_out[vrf]['vrfId']
        for key in json_out[vrf]:
            if key == 'vrfName' or key == 'vrfId':
                continue

            session = json_out[vrf][key]

            bgp_nbr = BgpSession()
            # Quagga/FRR names the default VRF 'Default' whereas Linux names it
            # nothing, and in netq we call it 'default'. Fix this.
            if vrf == 'Default':
                bgp_nbr.vrf = 'default'
            else:
                bgp_nbr.vrf = vrf.encode('ascii', 'ignore')
            bgp_nbr.vrfid = vrf_id
            bgp_nbr.peer_name = key.encode('ascii', 'ignore')
            bgp_nbr.state = session['bgpState'].encode('ascii', 'ignore')
            bgp_nbr.peer_router_id = (
                session['remoteRouterId'].encode('ascii', 'ignore')
            )
            bgp_nbr.peer_asn = session['remoteAs']
            bgp_nbr.peer_hostname = session.get('hostname', '').encode('ascii',
                                                                       'ignore')
            bgp_nbr.asn = int(session['localAs'])
            bgp_nbr.reason = 'N/A'
            if bgp_nbr.state == 'Established':
                bgp_nbr.up_time = float(session.get('bgpTimerUpEstablishedEpoch', 0.0))
            else:
                bgp_nbr.up_time = 0.0
                retcode = ServiceStatus.ERROR
            bgp_nbr.upd8_rx = session['messageStats']['updatesRecv']
            bgp_nbr.upd8_tx = session['messageStats']['updatesSent']
            afi_session = session['addressFamilyInfo']
            if 'lastResetDueTo' in afi_session:
                bgp_nbr.reason = (
                    afi_session['lastResetDueTo'].encode('ascii', 'ignore')
                )
                bgp_nbr.last_reset_time = \
                    float(afi_session['lastResetTimerMsecs'] / 1000)
            elif 'lastResetDueTo' in session:
                bgp_nbr.reason = (
                    session['lastResetDueTo'].encode('ascii', 'ignore')
                )
                if bgp_nbr.reason.startswith("BGP Notification"):
                    if 'lastNotificationReason' in session:
                        bgp_nbr.reason = (
                            session['lastNotificationReason'].encode('ascii',
                                                                     'ignore'))
                bgp_nbr.last_reset_time = \
                    float(session['lastResetTimerMsecs'] / 1000)
            else:
                bgp_nbr.last_reset_time = 0.0
            if 'IPv4 Unicast' in session['addressFamilyInfo']:
                bgp_nbr.ipv4_pfx_rcvd = (
                    int(afi_session['IPv4 Unicast']['acceptedPrefixCounter'])
                )
            else:
                bgp_nbr.ipv4_pfx_rcvd = 0
            if 'IPv6 Unicast' in session['addressFamilyInfo']:
                bgp_nbr.ipv6_pfx_rcvd = (
                    int(afi_session['IPv6 Unicast']['acceptedPrefixCounter'])
                )
            else:
                bgp_nbr.ipv6_pfx_rcvd = 0
            if 'connectionsDropped' in afi_session:
                bgp_nbr.conn_dropped = afi_session['connectionsDropped']
                bgp_nbr.conn_estd = afi_session['connectionsEstablished']
            elif 'connectionsDropped' in session:
                bgp_nbr.conn_dropped = session['connectionsDropped']
                bgp_nbr.conn_estd = session['connectionsEstablished']
            bgp_nbr.save()

    return retcode
